"""
File defining a class for a base machine learning model.
"""

from abc import ABC, abstractmethod
from typing import Self

import numpy as np


class AbstractModel(ABC):
    """Abstract class to implement a base machine learning model."""

    def __init__(self) -> None:
        super().__init__()
        self.X = np.array([])
        self.y = np.array([])

    @abstractmethod
    def fit(self, X: np.ndarray, y: np.ndarray) -> Self:
        """Fit the model.

        Args:
            X (np.ndarray): Inputs
            y (np.ndarray): Output

        Returns:
            Self: Fitted model.
        """

    @abstractmethod
    def predict(self, X: np.ndarray) -> np.ndarray:
        """Predict values with the model for the inputs given on argument.

        Args:
            X (np.ndarray): Inputs

        Returns:
            np.ndarray: Predicted values.
        """

    @abstractmethod
    def score(self, X: np.ndarray, y: np.ndarray) -> float:
        """Returns the score of the model for the
        inputs and output given on arguments.

        Args:
            X (np.ndarray): Inputs
            y (np.ndarray): Output

        Returns:
            float: Model score
        """

    def _reshape_ndarray(self, ndarray: np.ndarray) -> np.ndarray:
        """Reshape 1-D ndarray to ensure it has 2D shape.

        Args:
            ndarray (np.ndarray): ndarray to rehasepe

        Returns:
            np.ndarray: Reshaped ndarray
        """
        return (
            ndarray.reshape((ndarray.shape[0], 1))
            if len(ndarray.shape) == 1
            else ndarray
        )



"""
File defining an abstract regressor (for predicting continous `y` values)
"""

import numpy as np

from cmnemoi_learn.abstract_model import AbstractModel


class AbstractRegressor(AbstractModel):
    """Class defining an abstract regressor (for predicting continous `y` values).

    Args:
        AbstractModel (AbstractModel): Class defining an abstract model.
    """

    def score(self, X: np.ndarray, y: np.ndarray) -> float:
        """Returns the mean squared error of the model for the
        inputs and output given on arguments.

        Args:
            X (np.ndarray): Inputs
            y (np.ndarray): Output

        Returns:
            float: Mean Squared Error
        """
        y_pred = self.predict(X)
        return np.mean((y - y_pred) ** 2)



"""
File defining a logistic regression model.
"""
from typing import Optional, Self
import numpy as np

from cmnemoi_learn.classification.abstract_classifier import AbstractClassifier


class LogisticRegression(AbstractClassifier):
    """Logistic Regression model.

    The logistic regression tries to compute the
    probability a sample x to belong to the
    `Y=1` positive class : `P(Y=1|X=x)`.

    For this, it assumes `P(Y=1|X=x) = z(X.weights)`
    where z is the logistic function `z(x) = (1 + exp(-x))**-1`
    """

    def __init__(self, n_iter: int = 500, random_state: Optional[int] = None) -> None:
        super().__init__()
        self.n_iter = n_iter
        self.rng = np.random.default_rng(seed=random_state)
        self.weights = np.array([])

    def fit(self, X: np.ndarray, y: np.ndarray) -> Self:
        """Fit the model.

        The `weights` of the model are found
        by minimizing the negative log-likelihood of the problem
        (aka. log loss or cross entropy) :
        `−∑ ylog(z(X.weights)) + (1-y)log(1-z(X.weights))` with
        gradient descent algorithm.

        Args:
            X (np.ndarray): Inputs
            y (np.ndarray): Output

        Returns:
            Self: Fitted model.
        """
        self.X = self._get_inputs_with_bias_column(X)
        self.y = self._reshape_ndarray(y)
        self.weights = self._gradient_descent(n_iter=self.n_iter)
        return self

    def predict(self, X: np.ndarray) -> np.ndarray:
        """Predict the class of the inputs given on argument.

        Args:
            X (np.ndarray): Inputs

        Returns:
            np.ndarray: Output
        """
        X = self._get_inputs_with_bias_column(X)
        probabilities = self._logistic_function(X @ self.weights)
        return np.array(
            [1 if probability > self.threshold else 0 for probability in probabilities]
        )

    def _gradient_descent(self, n_iter: int, learning_rate: float = 0.5) -> np.ndarray:
        """Gradient descent algorithm.
        Find the best `weights` for logistic regression by minimizing the log loss.

        Args:
            n_iter (int, optional): Number of iterations.
            learning_rate (float, optional): Learning rate. Defaults to 0.5.

        Returns:
            Tuple[np.ndarray, float]: Weights and bias of the model.
        """
        n = self.X.shape[0]
        weights = self.rng.random((self.X.shape[1], 1))
        weights_gradient = np.array([])
        for _ in range(n_iter):
            weights_gradient = (
                (1 / n)
                * self.X.T
                @ (self._logistic_function(self.X @ weights) - self.y)
            )
            weights = weights - learning_rate * weights_gradient

        return weights

    def _get_inputs_with_bias_column(self, X: np.ndarray) -> np.ndarray:
        """Returns the inputs `X` with a `1`-filled bias column.

        Args:
            X (np.ndarray): Model inputs

        Returns:
            np.ndarray: New inputs with a bias column.
        """
        number_of_rows = X.shape[0]
        bias_column = np.ones((number_of_rows, 1))
        return np.hstack((bias_column, X))

    def _logistic_function(self, X: np.ndarray) -> np.ndarray:
        """Logistic function.
        `z(x) = (1 + exp(-x))**-1`

        Args:
            X (np.ndarray): Inputs

        Returns:
            np.ndarray: Outputs
        """
        return (1 + np.exp(-X)) ** -1



"""
File defining a Linear Regression model.
"""

from typing import Self
import numpy as np
from numpy.linalg import pinv

from cmnemoi_learn.regression.abstract_regressor import AbstractRegressor


class LinearRegression(AbstractRegressor):
    """
    Linear Regression model.
    `y = X.theta` where `theta` are the parameters of the model.
    """

    def __init__(self) -> None:
        super().__init__()
        self.theta = np.array([])

    def fit(self, X: np.ndarray, y: np.ndarray) -> Self:
        """Fit the Linear Regression model with normal equations solution.

        The optimal parameters `theta` of the model are the ones which minimize
        Residuals Sum of Squares : `RSS = Sum(y - X.theta)**2`.

        Args:
            X (np.ndarray): Inputs
            y (np.ndarray): Output

        Returns:
            LinearRegression: Fitted Linear Regression model.
        """
        self.X = self._get_inputs_with_bias_column(X)
        self.y = y

        self.theta = pinv(self.X.T @ self.X) @ (self.X.T @ self.y)

        return self

    def predict(self, X: np.ndarray) -> np.ndarray:
        """Predict new values with the Linear Regression model for the inputs given on argument.

        Args:
            X (np.ndarray): New inputs on which to predict.

        Returns:
            LinearRegression: Linear Regression model used to predict.
        """
        X = self._get_inputs_with_bias_column(X)
        return X @ self.theta

    def _get_inputs_with_bias_column(self, X: np.ndarray) -> np.ndarray:
        """Returns the inputs `X` with a `1`-filled bias column.

        Args:
            X (np.ndarray): Model inputs

        Returns:
            np.ndarray: New inputs with a bias column.
        """
        number_of_rows = X.shape[0]
        bias_column = np.ones((number_of_rows, 1))
        return np.hstack((bias_column, X))



"""
File defining an abstract classifier (for predicting discrete `y` values)
"""

import numpy as np

from cmnemoi_learn.abstract_model import AbstractModel


class AbstractClassifier(AbstractModel):
    """Class defining an abstract classifier (for predicting discrete `y` values).

    Args:
        AbstractModel (AbstractModel): Class defining an abstract model.
    """

    def __init__(self) -> None:
        super().__init__()
        self.threshold = 0.5

    def score(self, X: np.ndarray, y: np.ndarray) -> float:
        """Returns the accuracy of the model for the
        inputs and output given on arguments.

        Args:
            X (np.ndarray): Inputs
            y (np.ndarray): Output

        Returns:
            float: Accuracy
        """
        y_pred = self.predict(X)
        return float(
            np.mean(
                [
                    1 if yi_pred == yi_true else 0
                    for (yi_pred, yi_true) in zip(y_pred, y)
                ]
            )
        )
